/*

--------------------------------------------------------------------
notercycle - realtime networked staff notation tool for Processing

                  (c) Nick Fells 2014
                  
notercycle requires the oscP5 library from Andreas Schlegel:
http://www.sojamo.de/libraries/oscP5
You need to import this into Processing.

and uses Gonville font from Simon Tatham:
http://www.chiark.greenend.org.uk/~sgtatham/gonville
This is included with the notercycle distribution.

Notercycle is tested with Processing 2.1.1

The script is currently configured for string quartet, running
on two laptops across a network. One instance of the script should
be run with the 'two treble clef' settings below, the other (on the 
other laptop) with those commented out and the 'alto and bass clef' 
settings used instead. Reconfiguring for other ensembles/
combinations will require substantial tweaking of variables 
numstaves, npoffset, octoffset, maxpitch, minpitch, ottavapitch and
the size/layout variables listed below.

The script is designed for my piece Ps[c]yched for string quartet,
bikes and laptops, and employs a cueing system particular to that
piece. This will be explained in a short user manual at some point!

This work is licensed under a 
Creative Commons Attribution-NonCommercial 4.0 International License
which means it can be adapted and shared but not used for commercial 
purposes.

--------------------------------------------------------------------

*/

import oscP5.*;
import netP5.*;
 
int treble=0, alto=1, bass=2;
int numstaves=2;

/* 
IP number of remote MAX machine: currently set to receive from 
ALL IP addresses, which avoids having to set/get the IP address of 
the MAX machine (which will vary), but it does assume the three
machines are running on a private network.
*/

String maxmspIPAddress = "255.255.255.255";

// initialise staff/clef arrays (uncomment as necessary)

/*

// For two treble clef staves (aka first and second fiddle):

    int inPort = 12000; // data received from MAX on 12000
    int outPort = 13000; // end of line triggers etc sent on 13000
    int[] staves = {treble,treble}; 
    int[] npoffset = {0,0};  // noteposition offset
    int[] octoffset = {5,5};  // octave position offset
    int[] maxpitch = {101,101};  // max pitch to be displayed (higher pitches ignored)
    int[] minpitch = {55,55};    // min pitch to be displayed (lower pitches ignored)
    int[] ottavapitch = {96,96}; // high point above which ottava sign is used
*/

// For alto and bass clef staves (aka violas):

    int inPort = 12001;   // data received from MAX on 12001
    int outPort = 13001; // end of line triggers etc sent on 13001
    int[] staves = {alto,bass};
    int[] npoffset = {6,5};
    int[] octoffset = {5,4};
    int[] maxpitch = {93,84};
    int[] minpitch = {48,36};
    int[] ottavapitch = {84,68}; 


// NOTHING BELOW HERE SHOULD NEED TO BE CHANGED 
///////////////////////////////////////////////




// THESE GLOBALS SHOULDN'T NEED TO BE CHANGED

OscP5 oscP5;
NetAddress maxMsp;

PFont noteFont, typeFont;


int linespace, ledglen, ledglongfac, thickness;
int globalcue=0, flash=0, flashDelay=15, flashCount=0;
int flashFlag=0,flashNum=9;


int[] playheadx = {0,0};
int[] staffy = {0,0};

int[] notepos= {6,6,6,6,5,5,5,5,4,4,3,3,3,3,2,2,2,2,1,1,1,1,0,0};
int[] accidentals = {0,1,2,3,0,1,2,3,0,1,0,1,2,3,0,1,2,3,0,1,2,3,0,1};
String[] sharps = {"\uE16D","\uE170","\uE176"};



// VARIABLES GOVERNING SIZE/LAYOUT
int fontsize=60;              // notation font size
int textfontsize=fontsize/2;  // text font size;
int topstaffspacing=150;       // top margin
int staffspacing=400;         // spacing between staves
int startx=150;               // starting x position for playhead (100)
int playspacing=60;           // x advancement amount in pixels
int rhxlim=50;                // limit of x in from end of line before refresh
int clearledgerstop = 7;     // max no. of ledger lines above staff, for clearing purposes
int clearledgersbot = 3;


// relationship between font size and linespacing: 90/23; 45/12: linespacing = (0.25 * fontsize) + 1
// relationship between font size and ledger line size: ledger line len = fontsize/2; divided at ratio 1/5 (x-) to 4/5 (x+)
// fontsize ideally multiple of 2 and 7.5 (e.g. 30, 60, 90)


void setup() {
  
  int n;

// start listening for incoming OSC messages on port 12000:
  oscP5 = new OscP5(this, inPort);

// font parameters:  
  fill(0, 0, 0); // black is font colour
  noteFont = createFont("gonville-23.otf", fontsize, true); // font with smoothing on, 90 point
  typeFont = createFont("Helvetica", textfontsize, true);
  textFont(noteFont);
  textAlign(LEFT, BASELINE); 

// for testing font size and positioning: 
/*  line(0, 20, 200, 20);
  text("\uE118"+"\uE13E"+"\uE170", 20, 100);
*/

// positioning parameters:  

  linespace = int((0.25 * fontsize) + 1); //spacing between staff lines
  ledglen = int(fontsize/2);  // ledger line length
  ledglongfac = int(fontsize/7.5);  // long ledger line multiplier for semibreve
  thickness = 1;  // staff and ledger line thickness

// create window and staff:  
  size(displayWidth, displayHeight);
  background(220);
  for(n=0; n<numstaves; ++n) {
    staffy[n] = topstaffspacing + n*staffspacing;
    clearStaff(n);
    playheadx[n] = startx; // set current position of playhead to starting value
  }

// register plug method, clearStaff(), for '/clear' OSC message  
  oscP5.plug(this,"clearStaff","/clear");
  oscP5.plug(this,"displayCue","/cue");
  oscP5.plug(this,"flashCue","/flash");
  oscP5.plug(this,"displayMsg","/msg");

// setup return OSC address for remote MaxMSP
  maxMsp = new NetAddress(maxmspIPAddress, outPort);
} 

/*boolean sketchFullScreen() {
  return true;
}*/


void draw() {
  
  if(flash == 1) {
   if(frameCount % flashDelay == 0) {
     flashCount++;

     if(flashCount < flashNum) {
       flashFlag = 1 - flashFlag; // toggle flag 0/1
       if(flashFlag==0) { // if flag = 0, print cue text in red
        fill(255,0,0);
        textFont(typeFont);
        textSize(textfontsize*2.5);
        text(globalcue, 10, staffy[0]-35);
        text(globalcue, 10, staffy[1]-35);
        textSize(textfontsize);
        fill(0);
       } else { // if flag = 1, blank out the  cue text
        fill(220);
        noStroke();
        rect(10, staffy[0]-30, 125, -60);
        rect(10, staffy[1]-30, 125, -60);
        fill(0);
        stroke(0);
       }
     } else {
       flashCount = 0;
       flash = 0;
     }
   }
  }
}



// IF OSCEVENT RECEIVED, CREATE NOTE
void oscEvent(OscMessage messageIn) {

  float dur, midinote, pitchclass; 
  int pitchheight, ypos, octave, accid, pitchix, ledglong, s;
  boolean ottava=false;
  
  OscMessage outMsg = new OscMessage("/endofline");
  
  String[] msg = splitTokens(messageIn.addrPattern(), "/");  // get staff number from token 1 of input message address (/1, /2 etc)
  
 
  if(match(msg[0], "clear") != null) return;
  
  s = int(msg[0]); // staves are numbered from 0

// if end of line, send end of line message back to Max and return
  if(playheadx[s]>width-rhxlim) {
    outMsg.add(s);
    oscP5.send(outMsg, maxMsp);
    return;
  }

  dur = messageIn.get(0).floatValue(); // get the duration
  midinote = messageIn.get(1).floatValue(); // get the note number 

  if(midinote<minpitch[s]) return;  // if pitch out of range, just return without rendering or updating, i.e. ignore the pitch
  if(midinote>maxpitch[s]) return;
  if(midinote>ottavapitch[s]) {  // if too many ledger lines, bring it down an octave and set ottava flag
    midinote = midinote - 12;
    ottava = true;
  }

// notehead parameters:
  octave = floor(midinote/12); // calculate octave
  pitchclass = midinote%12; // calculate pitch class
  pitchix = (int)(pitchclass*2); // index into pitch-related arrays
  pitchheight = (notepos[pitchix]-npoffset[s])-((octave-octoffset[s])*7); // height of note on staff
  
 
// calculate note's y position on the staff  
  ypos = staffy[s]+((pitchheight+4)*linespace/2);

// get and draw accidental, updating playhead x pos accordingly
  accid = accidentals[pitchix];
  if(accid>0) {
    playheadx[s] += (9+(accid*2)); // allow space before accidental, scaled to the type
                                // and therefore width of accidental (i.e. 3/4 sharp needs
                                // more space than 1/4 sharp)
                                
    accidental(accid, pitchheight, playheadx[s], s);
    playheadx[s] += (7+(accid*2)); // and allow space between accidental and notehead
  }

// render notehead
  textFont(noteFont);
  
  if(dur>1250) {
    text("\uE169", playheadx[s], ypos);  // semibreve
    ledglong = 1;  // semibreve is wider so needs longer ledger lines
  }
  else if(dur>400) {
    text("\uE13F", playheadx[s], ypos);  // minim
    ledglong = 0;
  }
  else {
    text("\uE13E", playheadx[s], ypos);  // crotchet
    ledglong = 0;
  }

// draw ledger lines if necessary
  if(pitchheight>5) {
    ledgerlineLo(pitchheight, playheadx[s], ledglong, s);
  }
  if(pitchheight<-5) {
    ledgerlineHi(pitchheight, playheadx[s], ledglong, s);
  }

// draw ottava sign if necessary
  if(ottava) ottavaSign(pitchheight, playheadx[s], s);

// increment playhead x pos  
  playheadx[s] += playspacing*ceil(dur/125);

}


// ACCIDENTALS

// (qs, sharp, 3qs)
void accidental(int accid, int pch, int x, int s) {
  int accidy;
  accidy = staffy[s]+((pch+4)*linespace/2);

  textFont(noteFont);
  textAlign(RIGHT, BASELINE);
  text(sharps[accid-1], x, accidy);
  textAlign(LEFT, BASELINE);
}



// LEDGER LINES

// ledgerlines above the staff
void ledgerlineHi(int pch, int x, int llong, int s){
  int numlines, l, ledgery;
  strokeWeight(thickness);
  
  numlines = round((float)(-pch-5)/2);
  for(l=0; l<numlines; ++l) {
    // calc y pos as staff pos minus (stafflines(0-1) plus l) * spacing
    ledgery = staffy[s] - ((1+l)*linespace);
    line(x-(0.2*ledglen), ledgery, x+(0.8*ledglen)+(ledglongfac*llong), ledgery);  // 1/5 length to left of notehead, 4/5 to right - and allow long line for semibreve
  }
}

//ledgerlines below the staff
void ledgerlineLo(int pch, int x, int llong, int s){
  int numlines, l, ledgery;
  strokeWeight(thickness);
  
  numlines = round((float)(pch-5)/2);
  for(l=0; l<numlines; ++l) {
    // calc y pos as staff pos plus (stafflines(0-4) plus l) * spacing
    ledgery = staffy[s] + ((5+l)*linespace);
    line(x-(0.2*ledglen), ledgery, x+(0.8*ledglen)+(ledglongfac*llong), ledgery);
  }
}


// OTTAVA SIGN

void ottavaSign(int pch, int x, int s) {
  int otty, numlines;
  strokeWeight(thickness);
  
  numlines = round((float)(-pch-5)/2);
  otty = staffy[s] - ((numlines+1) * linespace);
  
  textFont(typeFont);
  textSize(textfontsize/1.5);
  text("8va", x-7, otty-5);
  textSize(textfontsize);

}


// REFRESH STAFF

// clear window and redraw fresh staff
void clearStaff(int n) {
  int y = staffy[n] - clearledgerstop*linespace; // y pos of clear allows for ledger lines above staff
  int clearheight = (5+(clearledgerstop+clearledgersbot))*linespace; // height of clearance area allows for 2 x max ledger lines, plus 5 lines of staff
  
  fill(220);
  noStroke();
  rect(0, y, width, clearheight);
  fill(0);
  stroke(0);
  
  staff(n);
  clef(n);
  if(flash == 0) displayCue(globalcue); // redraw cue no. only if it's not flashing (else it draws black when it should be flashing red)
  playheadx[n]=startx;
}


// draw stafflines
void staff(int n) {
  strokeWeight(thickness);

  for(int y=0; y<5; ++y) {
    line(0,staffy[n]+(y*linespace),width,staffy[n]+(y*linespace));
  } 
}


void clef(int n){
  int clefline=3; // clef line counting down from TOP LINE of staff 
  String glyph="\uE118";
  
  switch(staves[n]){
    case 0:            // treble
      clefline=3;
      glyph="\uE118";
      break;
    case 1:            // alto
      clefline=2;
      glyph="\uE116";
      break;
    case 2:            // bass
      clefline=1;
      glyph="\uE117";
      break;
  }
  textFont(noteFont);  
  text(glyph, 20, staffy[n]+clefline*linespace);
}

void displayCue(int n) {
  globalcue = n;
  fill(220);
  noStroke();
  rect(10, staffy[0]-30, 125, -60);
  rect(10, staffy[1]-30, 125, -60);
  fill(0);
  stroke(0);
  
  textFont(typeFont);
  textSize(textfontsize*2.5);
  text(n, 10, staffy[0]-35);
  text(n, 10, staffy[1]-35);
  textSize(textfontsize);
}

void flashCue() {
  flash = 1;
}

void displayMsg(int n, String s) {

  fill(220);
  noStroke();
  rect(30, staffy[n]+((5+clearledgersbot)*linespace), 1500, 30); // clear the message rectangle
  fill(0);
  stroke(0);
  
  textFont(typeFont);
  textSize(textfontsize);
  text(s, 30, staffy[n]+150);  // display the new message
  textSize(textfontsize);
}

